import requests

from fleece.httperror import HTTPError

TOKEN_URL_FMT = 'https://identity.api.rackspacecloud.com/v2.0/tokens/{token}'


def authenticate():
    def wrap(fxn):
        """Return a decorated callable."""
        def wrapped(*args, **kwargs):
            """Validate token and return auth context."""
            if not kwargs.get('token'):
                raise HTTPError(status=401)
            userinfo = validate(kwargs['token'])
            if 'userinfo' in kwargs:
                kwargs['userinfo'] = userinfo
            return fxn(*args, **kwargs)
        return wrapped
    return wrap


def validate(token):
    """Validate token and return auth context."""
    token_url = TOKEN_URL_FMT.format(token=token)
    headers = {
        'x-auth-token': token,
        'accept': 'application/json',
    }
    resp = requests.get(token_url, headers=headers)

    if not resp.status_code == 200:
        raise HTTPError(status=401)
    return resp.json()



from __future__ import absolute_import

import datetime

from requests.structures import CaseInsensitiveDict


class Hasher(dict):
    # http://stackoverflow.com/a/3405143/190597
    def __missing__(self, key):
        value = self[key] = type(self)()
        return value


def format_event(event, context):
    output_context = {
        'aws_request_id': context.aws_request_id,
        'client_context': context.client_context,
        'function_name': context.function_name,
        'function_version': context.function_version,
        'get_remaining_time_in_millis': context.get_remaining_time_in_millis(),
        'invoked_function_arn': context.invoked_function_arn,
        'log_group_name': context.log_group_name,
        'log_stream_name': context.log_stream_name,
        'memory_limit_in_mb': context.memory_limit_in_mb,
    }

    hashed_event = Hasher(event)

    request = {
        'requested-at': datetime.datetime.utcnow().isoformat(),
        'context': output_context,
        'operation': hashed_event['operation'],
        'requestor': hashed_event['parameters']['requestor'],
        'body': hashed_event['parameters']['request']['body'],
        'path': hashed_event['parameters']['request']['path'],
        'querystring': hashed_event['parameters']['request']['querystring'],
        'header': CaseInsensitiveDict(
            hashed_event['parameters']['request']['header']),
        'gateway': hashed_event['parameters']['gateway'],
    }

    return request



# flake8: noqa

from __future__ import absolute_import

from fleece.handlers.connexion import *
import fleece.log

logger = fleece.log.get_logger(__name__)


logger.warning("fleece.connexion has been moved to fleece.handlers.connexion "
               "- please update as this will not work in future versions.")



from functools import wraps
import logging
import os
from random import random
import sys
import time

from six import string_types
import structlog

LOG_FORMAT = '%(message)s'
DEFAULT_STREAM = sys.stdout
WRAPPED_DICT_CLASS = structlog.threadlocal.wrap_dict(dict)
ENV_APIG_REQUEST_ID = '_FLEECE_APIG_REQUEST_ID'
ENV_LAMBDA_REQUEST_ID = '_FLEECE_LAMBDA_REQUEST_ID'


def clobber_root_handlers():
    [logging.root.removeHandler(handler) for handler in
     logging.root.handlers[:]]


class logme(object):
    """Log requests and responses"""

    def __init__(self, level=logging.DEBUG, logger=None):
        self.level = level
        if not logger:
            self.logger = logging.getLogger()
        else:
            self.logger = logger

    def __call__(self, func):

        def wrapped(*args, **kwargs):
            self.logger.log(self.level, "Entering %s", func.__name__)
            response = func(*args, **kwargs)
            func_response_name = "{0}_response".format(func.__name__)
            kwarg = {func_response_name: response}
            self.logger.log(self.level, "Exiting %s", func.__name__, **kwarg)
            return response

        return wrapped


class RetryHandler(logging.Handler):
    """A logging handler that wraps another handler and retries its emit
    method if it fails. Useful for handlers that connect to an external
    service over the network, such as CloudWatch.

    The wait between retries uses an exponential backoff algorithm with full
    jitter, as described in
    https://www.awsarchitectureblog.com/2015/03/backoff.html.

    :param handler the handler to wrap with retries.
    :param max_retries the maximum number of retries before giving up. The
                       default is 5 retries.
    :param backoff_base the sleep time before the first retry. This time
                        doubles after each retry. The default is 0.1s.
    :param backoff_cap the max sleep time before a retry. The default is 1s.
    :param ignore_errors if set to False, when all retries are exhausted, the
                         exception raised by the original log call is
                         re-raised. If set to True, the error is silently
                         ignored. The default is True.
    """
    def __init__(self, handler, max_retries=5, backoff_base=0.1,
                 backoff_cap=1, ignore_errors=True):
        super(RetryHandler, self).__init__()
        self.handler = handler
        self.max_retries = max_retries
        self.backoff_base = backoff_base
        self.backoff_cap = backoff_cap
        self.ignore_errors = ignore_errors

    def emit(self, record):
        try:
            return self.handler.emit(record)
        except Exception as e:
            exc = e

        sleep = self.backoff_base
        for i in range(self.max_retries):
            time.sleep(sleep * random())
            try:
                return self.handler.emit(record)
            except:
                pass
            sleep = min(self.backoff_cap, sleep * 2)

        if not self.ignore_errors:
            raise exc


def _has_streamhandler(logger, level=None, fmt=LOG_FORMAT,
                       stream=DEFAULT_STREAM):
    """Check the named logger for an appropriate existing StreamHandler.

    This only returns True if a StreamHandler that exaclty matches
    our specification is found. If other StreamHandlers are seen,
    we assume they were added for a different purpose.
    """
    # Ensure we are talking the same type of logging levels
    # if they passed in a string we need to convert it to a number
    if isinstance(level, string_types):
        level = logging.getLevelName(level)

    for handler in logger.handlers:
        if not isinstance(handler, logging.StreamHandler):
            continue
        if handler.stream is not stream:
            continue
        if handler.level != level:
            continue
        if not handler.formatter or handler.formatter._fmt != fmt:
            continue
        return True
    return False


def inject_request_ids_into_environment(func):
    """Decorator for the Lambda handler to inject request IDs for logging."""

    @wraps(func)
    def wrapper(event, context):
        # This might not always be an API Gateway event, so only log the
        # request ID, if it looks like to be coming from there.
        if 'requestContext' in event:
            os.environ[ENV_APIG_REQUEST_ID] = event['requestContext'].get(
                'requestId', 'N/A')
        os.environ[ENV_LAMBDA_REQUEST_ID] = context.aws_request_id
        return func(event, context)

    return wrapper


def add_request_ids_from_environment(logger, name, event_dict):
    """Custom processor adding request IDs to the log event, if available."""
    if ENV_APIG_REQUEST_ID in os.environ:
        event_dict['api_request_id'] = os.environ[ENV_APIG_REQUEST_ID]
    if ENV_LAMBDA_REQUEST_ID in os.environ:
        event_dict['lambda.request_id'] = os.environ[ENV_LAMBDA_REQUEST_ID]
    return event_dict


def _configure_logger(logger_factory=None, wrapper_class=None):

    if not logger_factory:
        logger_factory = structlog.stdlib.LoggerFactory()
    if not wrapper_class:
        wrapper_class = structlog.stdlib.BoundLogger

    structlog.configure(
        processors=[
            structlog.stdlib.filter_by_level,
            add_request_ids_from_environment,
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt='iso'),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.processors.JSONRenderer(sort_keys=True)
        ],
        context_class=WRAPPED_DICT_CLASS,
        logger_factory=logger_factory,
        wrapper_class=wrapper_class,
        cache_logger_on_first_use=True)


def setup_root_logger(level=logging.DEBUG, stream=DEFAULT_STREAM,
                      logger_factory=None):
    _configure_logger(logger_factory=logger_factory)
    clobber_root_handlers()
    root_logger = logging.root
    stream_handler = logging.StreamHandler(stream)
    stream_handler.setLevel(level)
    stream_handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT))
    root_logger.addHandler(stream_handler)
    root_logger.setLevel(level)


def get_logger(name=None, level=None, stream=DEFAULT_STREAM,
               clobber_root_handler=True, logger_factory=None,
               wrapper_class=None):
    """Configure and return a logger with structlog and stdlib."""
    _configure_logger(
        logger_factory=logger_factory,
        wrapper_class=wrapper_class)
    log = structlog.get_logger(name)
    root_logger = logging.root
    if log == root_logger:
        if not _has_streamhandler(root_logger, level=level, stream=stream):
            stream_handler = logging.StreamHandler(stream)
            stream_handler.setLevel(level)
            stream_handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT))
            root_logger.addHandler(stream_handler)
        else:
            if clobber_root_handler:
                for handler in root_logger.handlers:
                    handler.setFormatter(logging.Formatter(fmt=LOG_FORMAT))
    if level:
        log.setLevel(level)
    return log


def initial_trace_and_context_binds(logger, trace_id, lambda_context):
    """A helper to set up standard trace_id and lambda_context binds"""
    return logger.new(
        trace_id=trace_id,
        lambda_context={
            "function_name": lambda_context.function_name,
            "function_version": lambda_context.function_version,
            "invoked_function_arn": lambda_context.invoked_function_arn,
            "memory_limit_in_mb": lambda_context.memory_limit_in_mb,
            "aws_request_id": lambda_context.aws_request_id,
            "log_group_name": lambda_context.log_group_name,
            "log_stream_name": lambda_context.log_stream_name,
        }
    )


getLogger = get_logger



try:
    from BaseHTTPServer import BaseHTTPRequestHandler
except ImportError:
    from http.server import BaseHTTPRequestHandler

# import lzstring
# lz = lzstring.LZString()
# lz.decompressFromBase64(SECRET)
SECRET = ('FAAj4yrAKVogfQeAlCV9qIDQ0agHTLQxxKK76U0GEKZg'
          '4Dkl9YA9NADoQfeJQHFiC4gAPgCJJ4np07BZS8OMqyo4'
          'kaNDcABoXUpoHePpAAuIxb5YQZq+cItbYXQFpitGjjfNgQAA')


class HTTPError(Exception):

    default_status = 500

    def __init__(self, status=None, message=None):
        """Initialize class."""
        responses = BaseHTTPRequestHandler.responses

        # Add some additional responses that aren't included...
        responses[418] = ('I\'m a teapot', SECRET)
        responses[422] = ('Unprocessable Entity',
                          'The request was well-formed but was'
                          ' unable to be followed due to semantic errors')

        self.status_code = status or self.default_status

        # Don't explode if provided status_code isn't found.
        _message = responses.get(self.status_code, [''])
        error_message = "{0:d}: {1}".format(self.status_code, _message[0])
        if message:
            error_message = "{0} - {1}".format(error_message, message)

        super(HTTPError, self).__init__(error_message)



#!/usr/bin/env python
import argparse
import base64
import json
import os
import re
import subprocess
import sys

import boto3
import ruamel.yaml as yaml
import ruamel.yaml.comments
import six
if six.PY2:  # noqa
    from StringIO import StringIO
else:
    from io import StringIO

from fleece.cli.run import run

if six.PY2:
    input = raw_input  # noqa

PARAMETER_STORE_NAME = re.compile('/[a-zA-Z0-9_\\.\\-\\/]*$')


class AWSCredentialCache(object):
    def __init__(self, rs_username, rs_api_key, env_config):
        self.rs_username = rs_username
        self.rs_api_key = rs_api_key
        self.environments = run.get_config(env_config)['environments']
        self.rax_token = None
        self.tenant = None
        self.awscreds = {}

    def _get_rax_token(self):
        if self.rax_token is None:
            self.rax_token, self.tenant = run.get_rackspace_token(
                self.rs_username, self.rs_api_key)
        return self.rax_token, self.tenant

    def get_awscreds(self, environment):
        if environment not in self.awscreds:
            account = None
            for env in self.environments:
                if env['name'] == environment:
                    account = env['account']
                    break
            if account is None:
                raise ValueError('Environment "{}" is not known, add it to '
                                 'environments.yml file'.format(environment))
            token, tenant = self._get_rax_token()
            self.awscreds[environment] = run.get_aws_creds(
                account, tenant, token)
        return self.awscreds[environment]


STATE = {
    'awscreds': None,  # cache of aws credentials
    'stages': {}       # environment/key assignments for each stage
}


def _get_stage_data(stage, data=None):
    if not data:
        data = STATE['stages']
    data = run.get_stage_data(stage, data)
    if data is None:
        raise ValueError('No match for stage "{}"'.format(stage))
    return data


def _get_kms_key(stage):
    stage_data = _get_stage_data(stage)
    try:
        key = stage_data['key']
    except IndexError:
        raise ValueError('No key defined for stage "{}"'.format(stage))
    if key.startswith('alias/') or key.startswith('arn:'):
        return key
    return 'alias/' + key


def _get_environment(stage):
    stage_data = _get_stage_data(stage)
    try:
        return stage_data['environment']
    except IndexError:
        raise ValueError('No environment defined for stage "{}"'.format(stage))


def _encrypt_text(text, stage):
    key = _get_kms_key(stage)
    environment = _get_environment(stage)
    awscreds = STATE['awscreds'].get_awscreds(environment)
    kms = boto3.client('kms', aws_access_key_id=awscreds['accessKeyId'],
                       aws_secret_access_key=awscreds['secretAccessKey'],
                       aws_session_token=awscreds['sessionToken'])
    r = kms.encrypt(KeyId=key, Plaintext=text.encode('utf-8'))
    return base64.b64encode(r['CiphertextBlob']).decode('utf-8')


def _decrypt_text(text, stage):
    environment = _get_environment(stage)
    awscreds = STATE['awscreds'].get_awscreds(environment)
    kms = boto3.client('kms', aws_access_key_id=awscreds['accessKeyId'],
                       aws_secret_access_key=awscreds['secretAccessKey'],
                       aws_session_token=awscreds['sessionToken'])
    r = kms.decrypt(CiphertextBlob=base64.b64decode(text.encode('utf-8')))
    return r['Plaintext'].decode('utf-8')


def _encrypt_item(data, stage, key):
    if (isinstance(data, six.text_type) or isinstance(data, six.binary_type)) \
            and data.startswith(':encrypt:'):
        if not stage:
            sys.stderr.write('Warning: Key "{}" cannot be encrypted because '
                             'it does not belong to a stage\n'.format(key))
        else:
            data = ':decrypt:' + _encrypt_text(data[9:], stage)
    elif isinstance(data, (dict, ruamel.yaml.comments.CommentedMap)):
        per_stage = [k.startswith('+') for k in data]
        if any(per_stage):
            if not all(per_stage):
                raise ValueError('Keys "{}" have a mix of stage and non-stage '
                                 'variables'.format(', '.join(data.keys)))
            key_prefix = key + '.' if key else ''
            for k, v in data.items():
                data[k] = _encrypt_item(v, stage=k[1:], key=key_prefix + k)
        else:
            data = _encrypt_dict(data, stage=stage, key=key)
    elif isinstance(data, list):
        data = _encrypt_list(data, stage=stage, key=key)
    return data


def _encrypt_list(data, stage, key):
    return [_encrypt_item(v, stage=stage, key=key + '[]') for v in data]


def _encrypt_dict(data, stage=None, key=''):
    key_prefix = key + '.' if key else ''
    for k, v in data.items():
        data[k] = _encrypt_item(v, stage=stage, key=key_prefix + k)
    return data


def import_config(args, input_file=None):
    if not input_file:
        input_file = sys.stdin
    source = input_file.read().strip()
    if source[0] == '{':
        # JSON input
        config = json.loads(source)
    else:
        # YAML input
        config = yaml.round_trip_load(source)

    STATE['stages'] = config['stages']
    config['config'] = _encrypt_dict(config['config'])
    with open(args.config, 'wt') as f:
        if config:
            yaml.round_trip_dump(config, f)


def _decrypt_item(data, stage, key, render):
    if (isinstance(data, six.text_type) or isinstance(data, six.binary_type)) \
            and data.startswith(':decrypt:'):
        data = _decrypt_text(data[9:], stage)
        if not render or render == 'ssm':
            data = ':encrypt:' + data
    elif isinstance(data, (dict, ruamel.yaml.comments.CommentedMap)):
        if len(data) == 0:
            return data
        per_stage = [k.startswith('+') for k in data]
        if any(per_stage):
            if not all(per_stage):
                raise ValueError('Keys "{}" have a mix of stage and non-stage '
                                 'variables'.format(', '.join(data.keys)))
        if render:
            if per_stage[0]:
                stage_data = _get_stage_data(
                    stage, data={k[1:]: v for k, v in data.items()})
                if stage_data:
                    data = _decrypt_item(
                        data.get(stage, stage_data),
                        stage=stage, key=key, render=render)
                else:
                    raise ValueError('Key "{}" has no value for stage '
                                     '"{}"'.format(key, stage))
            else:
                data = _decrypt_dict(data, stage=stage, key=key, render=render)
        else:
            key_prefix = key + '.' if key else ''
            for k, v in data.items():
                data[k] = _decrypt_item(v, stage=k[1:], key=key_prefix + k,
                                        render=render)
            data = _decrypt_dict(data, stage=stage, key=key, render=render)
    elif isinstance(data, list):
        data = _decrypt_list(data, stage=stage, key=key, render=render)
    return data


def _decrypt_list(data, stage, key, render):
    return [_decrypt_item(v, stage=stage, key=key + '[]', render=render)
            for v in data]


def _decrypt_dict(data, stage=None, key='', render=False):
    key_prefix = key + '.' if key else ''
    for k, v in data.items():
        data[k] = _decrypt_item(v, stage=stage, key=key_prefix + k,
                                render=render)
    return data


def export_config(args, output_file=None):
    if not output_file:
        output_file = sys.stdout
    if os.path.exists(args.config):
        with open(args.config, 'rt') as f:
            config = yaml.round_trip_load(f.read())
        STATE['stages'] = config['stages']
        config['config'] = _decrypt_dict(config['config'])
    else:
        config = {
            'stages': {
                env['name']: {
                    'environment': env['name'],
                    'key': 'enter-key-name-here'
                } for env in STATE['awscreds'].environments
            },
            'config': {}}

    if args.json:
        output_file.write(json.dumps(config, indent=4))
    elif config:
        yaml.round_trip_dump(config, output_file)


def edit_config(args):
    filename = '.fleece_edit_tmp'
    skip_export = False

    if os.path.exists(filename):
        p = input('A previously interrupted edit session was found. Do you '
                  'want to (C)ontinue that session or (A)bort it? ')
        if p.lower() == 'a':
            os.unlink(filename)
        elif p.lower() == 'c':
            skip_export = True

    if not skip_export:
        with open(filename, 'wt') as fd:
            export_config(args, output_file=fd)

    subprocess.call(args.editor + ' ' + filename, shell=True)

    with open(filename, 'rt') as fd:
        import_config(args, input_file=fd)

    os.unlink(filename)


def _read_config_file(args):
    """Decrypt config file, returns a tuple with stages and config."""
    stage = args.stage
    with open(args.config, 'rt') as f:
        config = yaml.safe_load(f.read())
    STATE['stages'] = config['stages']
    config['config'] = _decrypt_item(config['config'], stage=stage, key='',
                                     render=True)
    return config['stages'], config['config']


def write_to_parameter_store(env, prefix, config, ssm_kms_key=None):
    environment = _get_environment(env)
    awscreds = STATE['awscreds'].get_awscreds(environment)

    if not prefix.startswith('/'):
        msg = ('Parameter store names must be fully qualified (start with a '
               'slash), so the given prefix "{}"" is invalid.'.format(prefix))
        raise ValueError(msg)

    def validate(name, value):
        if name.count('/') > 15:
            msg = ('Error writing name "{}": parameter store names allow for '
                   'no more than 15 levels of hierarchy.'.format(name))
            raise ValueError(msg)

        if not PARAMETER_STORE_NAME.match(name):
            msg = ('Error: invalid parameter name "{}". Parameter store names '
                   'may consist of only symbols and letters (a-zA-Z0-9_.-/)'
                   .format(name))
            raise ValueError(msg)

        if not isinstance(value, (six.string_types, dict)):
            msg = ('Error: all config values must be strings or dictionaries '
                   'to work with parameter store, can\'t handle {} of type {}'
                   .format(name, type(value)))
            raise ValueError(msg)
        elif isinstance(value, dict):
            for k, v in value.items():
                validate('{}/{}'.format(name, k), v)

    validate(prefix, config)

    sts = boto3.client("sts",
                       aws_access_key_id=awscreds['accessKeyId'],
                       aws_secret_access_key=awscreds['secretAccessKey'],
                       aws_session_token=awscreds['sessionToken'])
    account_id = sts.get_caller_identity()["Account"]

    print('Writing config with parameter store prefix {prefix} to AWS '
          'account {account_id}'.format(prefix=prefix, account_id=account_id))

    ssm = boto3.client('ssm',
                       aws_access_key_id=awscreds['accessKeyId'],
                       aws_secret_access_key=awscreds['secretAccessKey'],
                       aws_session_token=awscreds['sessionToken'])
    kms = boto3.client('kms',
                       aws_access_key_id=awscreds['accessKeyId'],
                       aws_secret_access_key=awscreds['secretAccessKey'],
                       aws_session_token=awscreds['sessionToken'])

    def put(name, value):
        if isinstance(value, dict):
            for k, v in value.items():
                put('{}/{}'.format(name, k), v)
        elif isinstance(value, six.string_types):
            ps_name = name
            print('Writing {}...'.format(ps_name))
            if ssm_kms_key is not None:
                # fetch the full keyid from the alias
                ssm_kms_key_id = kms.describe_key(
                    KeyId=ssm_kms_key
                )['KeyMetadata']['KeyId']
            else:
                ssm_kms_key_id = None
            kwargs = {}
            if ssm_kms_key_id:
                kwargs["KeyId"] = ssm_kms_key_id
            ssm.put_parameter(
                Name=ps_name,
                Value=value,
                Type='SecureString',
                Overwrite=True,
                **kwargs
            )

    put(prefix, config)


def render_config(args, output_file=None):
    if not output_file:
        output_file = sys.stdout

    stages, config = _read_config_file(args)

    env = args.environment or args.stage

    if args.parameter_store is not None:
        return write_to_parameter_store(
            env=args.environment or args.stage,
            prefix=args.parameter_store,
            config=config,
            ssm_kms_key=args.ssm_kms_key,
        )

    if args.json or args.encrypt or args.python:
        rendered_config = json.dumps(
            config, indent=None if args.encrypt else 4,
            separators=(',', ':') if args.encrypt else (',', ': '))
    else:
        buf = StringIO()
        yaml.round_trip_dump(config, buf)
        rendered_config = buf.getvalue()
    if args.encrypt or args.python:
        STATE['stages'] = stages
        encrypted_config = []
        while rendered_config:
            buffer = _encrypt_text(rendered_config[:4096], env)
            rendered_config = rendered_config[4096:]
            encrypted_config.append(buffer)

        if not args.python:
            rendered_config = json.dumps(encrypted_config)
        else:
            rendered_config = '''ENCRYPTED_CONFIG = {}
import base64
import boto3
import json

def load_config():
    config_json = ''
    kms = boto3.client('kms')
    for buffer in ENCRYPTED_CONFIG:
        r = kms.decrypt(CiphertextBlob=base64.b64decode(buffer.encode(
            'utf-8')))
        config_json += r['Plaintext'].decode('utf-8')
    return json.loads(config_json)

CONFIG = load_config()
'''.format(encrypted_config)
    output_file.write(rendered_config)


def parse_args(args):
    parser = argparse.ArgumentParser(
        prog='fleece config',
        description=('Configuration management')
    )
    parser.add_argument('--config', '-c', default='config.yml',
                        help='Config file (default is config.yml)')
    parser.add_argument('--username', '-u', type=str,
                        default=os.environ.get('RS_USERNAME'),
                        help=('Rackspace username. Can also be set via '
                              'RS_USERNAME environment variable'))
    parser.add_argument('--apikey', '-k', type=str,
                        default=os.environ.get('RS_API_KEY'),
                        help=('Rackspace API key. Can also be set via '
                              'RS_API_KEY environment variable'))
    parser.add_argument('--environments', '-e', type=str,
                        default='./environments.yml',
                        help=('Path to YAML config file with defined accounts '
                              'and environment names. Defaults to '
                              './environments.yml'))
    subparsers = parser.add_subparsers(help='Sub-command help')

    import_parser = subparsers.add_parser(
        'import', help='Import configuration from stdin')
    import_parser.set_defaults(func=import_config)

    export_parser = subparsers.add_parser(
        'export', help='Export configuration to stdout')
    export_parser.add_argument('--json', action='store_true',
                               help='Use JSON format (default is YAML)')
    export_parser.set_defaults(func=export_config)

    edit_parser = subparsers.add_parser('edit', help='Edit configuration')
    edit_parser.add_argument('--json', action='store_true',
                             help='Use JSON format (default is YAML)')
    edit_parser.add_argument(
        '--editor', '-e', default=os.environ.get('FLEECE_EDITOR', 'vi'),
        help='Text editor (defaults to $FLEECE_EDITOR, or else "vi")')
    edit_parser.set_defaults(func=edit_config)

    render_parser = subparsers.add_parser(
        'render', help='Render configuration for an environment')
    render_parser.add_argument('--environment', '-e',
                               help=('Environment name (default is the '
                                     'stage name'))
    render_parser.add_argument('--json', action='store_true',
                               help='Use JSON format (default is YAML)')
    render_parser.add_argument('--encrypt', action='store_true',
                               help='Encrypt rendered configuration')
    render_parser.add_argument('--python', action='store_true',
                               help=('Generate Python module with encrypted '
                                     'configuration'))
    render_parser.add_argument('--parameter-store', type=str, default=None,
                               help=('Write configuration to AWS '
                                     'parameter-store using the given prefix '
                                     'for the selected stage\'s environment'))
    render_parser.add_argument(
        '--ssm-kms-key',
        type=str,
        default=None,
        help=(
            'KMS key ID or alias to use for encrypting config in SSM. Use with'
            ' `--parameter-store`.'
        ),
    )
    render_parser.add_argument('stage', help='Target stage name')
    render_parser.set_defaults(func=render_config)

    return parser.parse_args(args)


def main(args):
    parsed_args = parse_args(args)

    STATE['awscreds'] = AWSCredentialCache(rs_username=parsed_args.username,
                                           rs_api_key=parsed_args.apikey,
                                           env_config=parsed_args.environments)
    parsed_args.func(parsed_args)



"""Fleece package attributes and metadata."""

__all__ = (
    '__title__',
    '__summary__',
    '__author__',
    '__email__',
    '__license__',
    '__version__',
    '__copyright__',
    '__url__',
)

__title__ = 'fleece'
__summary__ = 'Wrap the lamb...da'
__author__ = 'Rackers'
__email__ = 'bruce.stringer@rackspace.com'
__version__ = '0.19.0'
__license__ = 'Apache License, Version 2.0'
__keywords__ = ['fleece', 'lambda']
__copyright__ = 'Copyright Rackspace US, Inc. 2016'
__url__ = 'https://github.com/racker/fleece'



from __future__ import absolute_import

try:
    from cStringIO import cStringIO as StringIO
except ImportError:
    # Python 3
    from io import StringIO
import json
import os.path
try:
    from urllib import urlencode
except ImportError:
    # Python 3
    from urllib.parse import urlencode

import connexion
import six
import werkzeug.wrappers

from fleece.handlers.wsgi import build_wsgi_environ_from_event
from fleece import httperror
import fleece.log

_app_cache = {}

RESPONSE_CONTRACT_VIOLATION = 'Response body does not conform to specification'


class FleeceApp(connexion.App):
    """Wrapper around `connexion.App` with added helpers for Lambda."""

    def __init__(self, *args, **kwargs):
        """Create a FleeceApp.

        Parameters are identical to a `connexion.App, with the exception of the
        below keyword argument.

        :param logging.Logger logger:
            A Logger instance returned by `fleece.log.get_logger()` to be used
            for capturing details about errors.

            If `logger` is None, a default logger object will be created.
        """
        logger = kwargs.pop('logger', None)
        if logger is None:
            self.logger = fleece.log.get_logger(__name__)
        else:
            self.logger = logger

        super(FleeceApp, self).__init__(*args, **kwargs)

        # Capture and log any unexpected exceptions raise by handler code:
        def error_handler(exception):
            self.logger.exception(exception)
            # A `None` return value will not modify the response.
            # It's possible to return new types of responses from an error
            # handler.
            # To do so, simply return a new `werkzeug.wrappers.Response`
            # object.

        self.add_error_handler(Exception, error_handler)

    def call_api(self, event):
        """Make a request against the API defined by this app.

        Return any result from the API call as a JSON object/dict. In the event
        of a client or server error response from the API endpoint handler (4xx
        or 5xx), raise a :class:`fleece.httperror.HTTPError` containing some
        context about the error. Any unexpected exception encountered while
        executing an API endpoint handler will be appropriately raised as a
        generic 500 error with no error context (in order to not expose too
        much of the internals of the application).

        :param dict event:
            Dictionary containing the entire request payload passed to the
            Lambda function handler.
        :returns:
            JSON object (as a `list` or `dict`) containing the response data
            from the API endpoint.
        :raises:
            :class:`fleece.httperror.HTTPError` on 4xx or 5xx responses from
            the API endpoint handler, or a 500 response on an unexpected
            failure (due to a bug in handler code, for example).
        """
        try:
            environ = _build_wsgi_env(event, self.import_name)
            response = werkzeug.wrappers.Response.from_app(self, environ)
            response_dict = json.loads(response.get_data())

            if 400 <= response.status_code < 500:
                if ('error' in response_dict and
                        'message' in response_dict['error']):
                    # FIXME(larsbutler): If 'error' is not a collection
                    # (list/dict) and is a scalar such as an int, the check
                    # `message in response_dict['error']` will blow up and
                    # result in a generic 500 error. This check assumes too
                    # much about the format of error responses given by the
                    # handler. It might be good to allow this handling to
                    # support custom structures.
                    msg = response_dict['error']['message']
                elif 'detail' in response_dict:
                    # Likely this is a generated 400 response from Connexion.
                    # Reveal the 'detail' of the message to the user.
                    # NOTE(larsbutler): If your API handler explicitly returns
                    # a 'detail' key in in the response, be aware that this
                    # will be exposed to the client.
                    msg = response_dict['detail']
                else:
                    # Respond with a generic client error.
                    msg = 'Client Error'
                # FIXME(larsbutler): The logic above still assumes a lot about
                # the API response. That's not great. It would be nice to make
                # this more flexible and explicit.
                self.logger.error(
                    'Raising 4xx error',
                    http_status=response.status_code,
                    message=msg,
                )
                raise httperror.HTTPError(
                    status=response.status_code,
                    message=msg,
                )
            elif 500 <= response.status_code < 600:
                if response_dict['title'] == RESPONSE_CONTRACT_VIOLATION:
                    # This case is generally enountered if the API endpoint
                    # handler code does not conform to the contract dictated by
                    # the Swagger specification.
                    self.logger.error(
                        RESPONSE_CONTRACT_VIOLATION,
                        detail=response_dict['detail']
                    )
                else:
                    # This case will trigger if
                    # a) the handler code raises an unexpected exception
                    # or
                    # b) the handler code explicitly returns a 5xx error.
                    self.logger.error(
                        'Raising 5xx error',
                        response=response_dict,
                        http_status=response.status_code,
                    )
                raise httperror.HTTPError(status=response.status_code)
            else:
                return response_dict
        except httperror.HTTPError:
            self.logger.exception('HTTPError')
            raise
        except Exception:
            self.logger.exception('Unhandled exception')
            raise httperror.HTTPError(status=500)

    def call_proxy_api(self, event):
        """Make a request against the API defined by this app.

        Return any result from the API call as a dict that is expected by the
        AWS_PROXY type integration in API Gateway.

        :param dict event:
            Dictionary containing the entire request payload passed to the
            Lambda function handler.
        :returns:
            Dictionary containing the response data from the API endpoint.
        :raises:
            :class:`fleece.httperror.HTTPError` on 4xx or 5xx responses from
            the API endpoint handler, or a 500 response on an unexpected
            failure (due to a bug in handler code, for example).
        """
        try:
            environ = build_wsgi_environ_from_event(event)
            response = werkzeug.wrappers.Response.from_app(self, environ)
            return {
                'statusCode': response.status_code,
                'headers': {
                    header: value for header, value in response.headers.items()
                },
                'body': response.get_data(as_text=True),
            }
        except Exception:
            # We should actually never get here, because exceptions raised
            # within the application are handled by the error handlers, and the
            # default one will return a clean 500 error during the happy path
            # above.
            self.logger.exception('Unhandled exception')
            return {
                'statusCode': 500,
                'headers': {},
                'body': json.dumps({
                    'error': {
                        'message': 'Internal server error.',
                    },
                }),
            }


def _build_wsgi_env(event, app_name):
    """Turn the Lambda/API Gateway request event into a WSGI environment dict.

    :param dict event:
        The event parameters passed to the Lambda function entrypoint.
    :param str app_name:
        Name of the API application.
    """
    gateway = event['parameters']['gateway']
    request = event['parameters']['request']
    ctx = event['rawContext']
    headers = request['header']
    body = six.text_type(json.dumps(request['body']))

    # Render the path correctly so connexion/flask will pass the path params to
    # the handler function correctly.
    # Basically, this replaces "/foo/{param1}/bar/{param2}" with
    # "/foo/123/bar/456".
    path = gateway['resource-path'].format(
        **event['parameters']['request']['path']
    )
    environ = {
        'PATH_INFO': path,
        'QUERY_STRING': urlencode(request['querystring']),
        'REMOTE_ADDR': ctx['identity']['sourceIp'],
        'REQUEST_METHOD': ctx['httpMethod'],
        'SCRIPT_NAME': app_name,
        'SERVER_NAME': app_name,
        'SERVER_PORT': headers.get('X-Forwarded-Port', '80'),
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': headers.get('X-Forwarded-Proto', 'http'),
        'wsgi.input': StringIO(body),
        'wsgi.errors': StringIO(),
        'wsgi.multiprocess': False,
        'wsgi.multithread': False,
        'wsgi.run_once': False,
        'CONTENT_TYPE': headers.get('Content-Type', 'application/json'),
    }
    if ctx['httpMethod'] in ['POST', 'PUT', 'PATCH']:
        environ['CONTENT_LENGTH'] = str(len(body))

    for header_name, header_value in headers.items():
        wsgi_name = 'HTTP_{}'.format(header_name.upper().replace('-', '_'))
        environ[wsgi_name] = str(header_value)

    return environ


def get_connexion_app(app_name, app_swagger_path, strict_validation=True,
                      validate_responses=True, cache_app=True, logger=None):
    # Optionally cache application instances, because it takes a significant
    # amount of time to process the Swagger definition, and we shouldn't be
    # doing it on every single request.
    if app_name not in _app_cache or not cache_app:
        full_path_to_swagger_yaml = os.path.abspath(app_swagger_path)
        app = FleeceApp(
            app_name,
            specification_dir=os.path.dirname(full_path_to_swagger_yaml),
            logger=logger,
        )
        app.add_api(
            os.path.basename(full_path_to_swagger_yaml),
            strict_validation=strict_validation,
            validate_responses=validate_responses,
        )
        _app_cache[app_name] = app

    return _app_cache[app_name]


def call_api(event, app_name, app_swagger_path, logger, strict_validation=True,
             validate_responses=True, cache_app=True):
    """Wire up the incoming Lambda/API Gateway request to an application.

    :param dict event:
        Dictionary containing the entire request template. This can vary wildly
        depending on the template structure and contents.
    :param str app_name:
        Name of the API application.
    :param str app_swagger_path:
        Local path to the Swagger API YAML file.
    :param logging.Logger logger:
        A Logger instance returned by `fleece.log.get_logger()` to be used for
        capturing details about errors.
    :param bool strict_validation:
        Toggle to enable/disable Connexion's parameter validation.
    :param bool validate_responses:
        Toggle to enable/disable Connexion's response validation.
    :param bool cache_app:
        Toggle to enable/disable the caching of the Connextion/Flask app
        instance. It's on by default, because it provides a significant
        performance improvement in the Lambda runtime environment.
    """
    app = get_connexion_app(
        app_name=app_name,
        app_swagger_path=app_swagger_path,
        strict_validation=strict_validation,
        validate_responses=validate_responses,
        cache_app=cache_app,
    )
    return app.call_api(event)


def call_proxy_api(event, app_name, app_swagger_path, logger,
                   strict_validation=True, validate_responses=True,
                   cache_app=True):
    """Wire up the incoming Lambda/API Gateway request to an application.

    Integration type of the resource must be AWS_LAMBDA in order for this to
    work.

    :param dict event:
        Dictionary containing the entire request template. This can vary wildly
        depending on the template structure and contents.
    :param str app_name:
        Name of the API application.
    :param str app_swagger_path:
        Local path to the Swagger API YAML file.
    :param logging.Logger logger:
        A Logger instance returned by `fleece.log.get_logger()` to be used for
        capturing details about errors.
    :param bool strict_validation:
        Toggle to enable/disable Connexion's parameter validation.
    :param bool validate_responses:
        Toggle to enable/disable Connexion's response validation.
    :param bool cache_app:
        Toggle to enable/disable the caching of the Connextion/Flask app
        instance. It's on by default, because it provides a significant
        performance improvement in the Lambda runtime environment.
    """
    app = get_connexion_app(
        app_name=app_name,
        app_swagger_path=app_swagger_path,
        strict_validation=strict_validation,
        validate_responses=validate_responses,
        cache_app=cache_app,
    )
    return app.call_proxy_api(event)



from werkzeug.test import EnvironBuilder


def build_wsgi_environ_from_event(event):
    """Create a WSGI environment from the proxy integration event."""
    params = event.get('queryStringParameters')
    environ = EnvironBuilder(method=event.get('httpMethod') or 'GET',
                             path=event.get('path') or '/',
                             headers=event.get('headers') or {},
                             data=event.get('body') or b'',
                             query_string=params or {}).get_environ()
    environ['SERVER_PORT'] = 443
    if 'execute-api' in environ['HTTP_HOST']:
        # this is the API-Gateway hostname, which takes the stage as the first
        # script path component
        environ['SCRIPT_NAME'] = '/' + event['requestContext'].get('stage')
    else:
        # we are using our own hostname, nothing gets added to the script path
        environ['SCRIPT_NAME'] = ''
    environ['wsgi.url_scheme'] = 'https'
    environ['lambda.event'] = event
    return environ


def wsgi_handler(event, context, app, logger):
    """lambda handler function.
    This function runs the WSGI app with it and collects its response, then
    translates the response back into the format expected by the API Gateway
    proxy integration.
    """

    environ = build_wsgi_environ_from_event(event)
    wsgi_status = []
    wsgi_headers = []

    logger.info('Processing {} request'.format(environ['REQUEST_METHOD']))

    def start_response(status, headers):
        if len(wsgi_status) or len(wsgi_headers):
            raise RuntimeError('start_response called more than once!')
        wsgi_status.append(status)
        wsgi_headers.append(headers)

    resp = list(app(environ, start_response))
    proxy = {'statusCode': int(wsgi_status[0].split()[0]),
             'headers': {h[0]: h[1] for h in wsgi_headers[0]},
             'body': b''.join(resp).decode('utf-8')}

    logger.info("Returning {}".format(proxy['statusCode']),
                http_status=proxy['statusCode'])

    return proxy



"""This module bridges the gap of missing Python support in the AWS X-Ray
Lambda integration.

After enabling active tracing under Advanced settings on the Configuration tab
of your Lambda function in the AWS Console (or using the
`update_function_configuration` API call), you can mark any function or method
for tracing by using the `@trace_xray_subsegment` decorator.

This module also provides wrappers for boto and requests so that any AWS API
call, or HTTP request will be automatically traced by X-Ray, but you have to
explicitly allow this behavior by calling `monkey_patch_botocore_for_xray`
and/or `monkey_patch_requests_for_xray`. The best place to do this would be the
main handler module where the Lambda entry point is defined.
"""
from collections import namedtuple
import json
import os
import socket
import string
import threading
import time
import uuid

from botocore.exceptions import ClientError
import wrapt

from fleece import log

LOGGER = log.get_logger('fleece.xray')

XRAY_DAEMON_HEADER = {'format': 'json', 'version': 1}

XRayDaemon = namedtuple('XRayDaemon', ['ip_address', 'port'])
XRayTraceID = namedtuple('XRayTraceID', ['trace_id', 'parent_id', 'sampled'])

ERROR_HANDLING_GENERIC = 'generic'
ERROR_HANDLING_BOTOCORE = 'botocore'

# Thread-local storage for parent IDs
threadlocal = threading.local()


class XRayDaemonNotFoundError(Exception):
    pass


class StringJSONEncoder(json.JSONEncoder):
    """Simple encoder that allows us to serialize everything into JSON."""

    def default(self, o):
        try:
            return super(StringJSONEncoder, self).default(o)
        except TypeError:
            return str(o)


def generate_subsegment_id():
    """Generate a random ID according to the X-Ray specs."""
    return uuid.uuid4().hex[:16]


def get_trace_id():
    """Parse X-Ray Trace ID environment variable.

    The value looks something like this:

        Root=1-5901e3bc-8da3814a5f3ccbc864b66ecc;Parent=328f72132deac0ce;Sampled=1

    `Root` is the main X-Ray Trace ID, `Parent` points to the top-level
    segment, and `Sampled` shows whether the current request should be traced
    or not.

    If the environment variable doesn't exist, just return an `XRayTraceID`
    instance with default values, which means that tracing will be skipped
    due to `sampled` being set to `False`.
    """
    raw_trace_id = os.environ.get('_X_AMZN_TRACE_ID', '')
    trace_id_parts = raw_trace_id.split(';')
    trace_kwargs = {
        'trace_id': None,
        'parent_id': None,
        'sampled': False,
    }
    if trace_id_parts[0] != '':
        # This means the trace ID environment variable is not empty
        for part in trace_id_parts:
            name, value = part.split('=')
            if name == 'Root':
                trace_kwargs['trace_id'] = value
            elif name == 'Parent':
                trace_kwargs['parent_id'] = value
            elif name == 'Sampled':
                trace_kwargs['sampled'] = bool(int(value))

    return XRayTraceID(**trace_kwargs)


def get_xray_daemon():
    """Parse X-Ray Daemon address environment variable.

    If the environment variable is not set, raise an exception to signal that
    we're unable to send data to X-Ray.
    """
    env_value = os.environ.get('AWS_XRAY_DAEMON_ADDRESS')
    if env_value is None:
        raise XRayDaemonNotFoundError()

    xray_ip, xray_port = env_value.split(':')
    return XRayDaemon(ip_address=xray_ip, port=int(xray_port))


def send_data_on_udp(ip_address, port, data):
    """Helper function to send a string over UDP to a specific IP/port."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        sock.sendto(data.encode('utf-8'), (ip_address, port))
    except:
        LOGGER.exception('Failed to send trace to X-Ray Daemon')
    finally:
        sock.close()


def send_segment_document_to_xray_daemon(segment_document):
    """Format and send document to the X-Ray Daemon."""
    try:
        xray_daemon = get_xray_daemon()
    except XRayDaemonNotFoundError:
        LOGGER.error('X-Ray Daemon not running, skipping send')
        return

    message = u'{header}\n{document}'.format(
        header=json.dumps(XRAY_DAEMON_HEADER),
        document=json.dumps(
            segment_document,
            ensure_ascii=False,
            cls=StringJSONEncoder,
        ),
    )

    send_data_on_udp(
        ip_address=xray_daemon.ip_address,
        port=xray_daemon.port,
        data=message,
    )


def send_subsegment_to_xray_daemon(subsegment_id, parent_id,
                                   start_time, end_time=None, name=None,
                                   namespace='remote', extra_data=None):
    """High level function to send data to the X-Ray Daemon.

    If `end_time` is set to `None` (which is the default), a partial subsegment
    will be sent to X-Ray, which has to be completed by a subsequent call to
    this function after the underlying operation finishes.

    The `namespace` argument can either be `aws` or `remote` (`local` is used
    as a default, but  it's not officially supported by X-Ray).

    The `extra_data` argument must be a `dict` that is used for updating the
    segment document with arbitrary data.
    """
    extra_data = extra_data or {}
    trace_id = get_trace_id()
    segment_document = {
        'type': 'subsegment',
        'id': subsegment_id,
        'trace_id': trace_id.trace_id,
        'parent_id': parent_id,
        'start_time': start_time,
    }
    if end_time is None:
        segment_document['in_progress'] = True
    else:
        segment_document.update({
            'end_time': end_time,
            'name': name,
            'namespace': namespace,
        })
        segment_document.update(extra_data)

    LOGGER.debug(
        'Prepared segment document for X-Ray Daemon',
        segment_document=segment_document,
    )
    send_segment_document_to_xray_daemon(segment_document)


def get_parent_id():
    """Retrieve the parent ID from the thread-local storage."""
    return getattr(threadlocal, 'parent_id', None)


def get_parent_id_from_trace_id():
    """Retrieve the parent ID from the trace ID environment variable."""
    trace_id = get_trace_id()
    return trace_id.parent_id


def set_parent_id(parent_id):
    """Store parent ID in the thread-local storage."""
    threadlocal.parent_id = parent_id


def get_function_name(wrapped, instance, args, kwargs):
    """Return the wrapped function's name."""
    return wrapped.__name__


def generic_xray_wrapper(wrapped, instance, args, kwargs, name, namespace,
                         metadata_extractor,
                         error_handling_type=ERROR_HANDLING_GENERIC):
    """Wrapper function around existing calls to send traces to X-Ray.

    `wrapped` is the original function, `instance` is the original instance,
    if the function is an instance method (otherwise it'll be `None`), `args`
    and `kwargs` are the positional and keyword arguments the original function
    was called with.

    The `name` argument can either be a `string` or a callable. In the latter
    case the function must accept the following arguments:

        def callback(wrapped, instance, args, kwargs):
            pass

    The `metadata_extractor` argument has to be a function with the following
    definition:

        def callback(wrapped, instance, args, kwargs, return_value):
            pass

    It has to return a `dict` that will be used to extend the segment document.

    The `error_handling_type` determines how exceptions raised by the wrapped
    function are handled. Currently `botocore` requires some special care.
    """
    if not get_trace_id().sampled:
        # Request not sampled by X-Ray, let's get to the call
        # immediately.
        LOGGER.debug('Request not sampled by X-Ray, skipping trace')
        return wrapped(*args, **kwargs)

    start_time = time.time()
    error = False
    cause = None

    # Fetch the parent ID from the current thread, and set it to the current
    # subsegment, so that downstream subsegments will be correctly nested.
    original_parent_id = get_parent_id()
    # If not parent ID exists in the thread-local storage, it means we're at
    # the topmost level, so we have to retrieve the parent ID from the trace ID
    # environment variable.
    parent_id = original_parent_id or get_parent_id_from_trace_id()
    subsegment_id = generate_subsegment_id()
    set_parent_id(subsegment_id)
    # Send partial subsegment to X-Ray, so that it'll know about the relations
    # upfront (otherwise we'll lose data, since downstream subsegments will
    # have invalid parent IDs).
    send_subsegment_to_xray_daemon(
        subsegment_id=subsegment_id,
        parent_id=parent_id,
        start_time=start_time,
    )
    try:
        return_value = wrapped(*args, **kwargs)
    except Exception as exc:
        error = True
        cause = {
            'exceptions': [
                {
                    'message': str(exc),
                    'type': '{}.{}'.format(
                        type(exc).__module__,
                        type(exc).__name__,
                    ),
                }
            ]
        }
        if error_handling_type == ERROR_HANDLING_GENERIC:
            return_value = None
        elif error_handling_type == ERROR_HANDLING_BOTOCORE:
            if isinstance(exc, ClientError):
                return_value = exc.response
            else:
                return_value = {}
        raise
    finally:
        end_time = time.time()
        extra_data = metadata_extractor(
            wrapped=wrapped,
            instance=instance,
            args=args,
            kwargs=kwargs,
            return_value=return_value,
        )
        extra_data['error'] = error
        if error:
            extra_data['cause'] = cause
        # We allow the name to be determined dynamically when a function is
        # passed in as the `name` argument.
        if callable(name):
            name = name(wrapped, instance, args, kwargs)
        send_subsegment_to_xray_daemon(
            subsegment_id=subsegment_id,
            parent_id=parent_id,
            start_time=start_time,
            end_time=end_time,
            name=name,
            namespace=namespace,
            extra_data=extra_data,
        )
        # After done with reporting the current subsegment, reset parent
        # ID to the original one.
        set_parent_id(original_parent_id)

    return return_value


def noop_function_metadata(wrapped, instance, args, kwargs, return_value):
    """Make sure that metadata is not changed."""
    return {}


def extract_function_metadata(wrapped, instance, args, kwargs, return_value):
    """Stash the `args` and `kwargs` into the metadata of the subsegment."""
    LOGGER.debug(
        'Extracting function call metadata',
        args=args,
        kwargs=kwargs,
    )
    return {
        'metadata': {
            'args': args,
            'kwargs': kwargs,
        },
    }


def trace_xray_subsegment(skip_args=False):
    """Can be applied to any function or method to be traced by X-Ray.

    If `skip_args` is True, the arguments of the function won't be sent to
    X-Ray.
    """
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        metadata_extractor = (
            noop_function_metadata
            if skip_args
            else extract_function_metadata
        )
        return generic_xray_wrapper(
            wrapped, instance, args, kwargs,
            name=get_function_name,
            namespace='local',
            metadata_extractor=metadata_extractor,
        )

    return wrapper


def get_service_name(wrapped, instance, args, kwargs):
    """Return the AWS service name the client is communicating with."""
    if 'serviceAbbreviation' not in instance._service_model.metadata:
        return instance._service_model.metadata['endpointPrefix']
    return instance._service_model.metadata['serviceAbbreviation']


def extract_aws_metadata(wrapped, instance, args, kwargs, return_value):
    """Provide AWS metadata for improved visualization.

    See documentation for this data structure:
    http://docs.aws.amazon.com/xray/latest/devguide/xray-api-segmentdocuments.html#api-segmentdocuments-aws
    """
    response = return_value
    LOGGER.debug(
        'Extracting AWS metadata',
        args=args,
        kwargs=kwargs,
    )
    if 'operation_name' in kwargs:
        operation_name = kwargs['operation_name']
    else:
        operation_name = args[0]

    # Most of the time the actual keyword arguments to the client call are
    # passed in as a positial argument after the operation name.
    if len(kwargs) == 0 and len(args) == 2:
        kwargs = args[1]

    region_name = instance._client_config.region_name

    response_metadata = response.get('ResponseMetadata')

    metadata = {
        'aws': {
            'operation': operation_name,
            'region': region_name,
        }
    }

    if 'TableName' in kwargs:
        metadata['aws']['table_name'] = kwargs['TableName']
    if 'QueueUrl' in kwargs:
        metadata['aws']['queue_url'] = kwargs['QueueUrl']

    if response_metadata is not None:
        metadata['http'] = {
            'response': {
                'status': response_metadata['HTTPStatusCode'],
            },
        }
        metadata['aws']['request_id'] = response_metadata['RequestId']

    return metadata


def xray_botocore_api_call(wrapped, instance, args, kwargs):
    """Wrapper around botocore's base client API call method."""
    return generic_xray_wrapper(
        wrapped, instance, args, kwargs,
        name=get_service_name,
        namespace='aws',
        metadata_extractor=extract_aws_metadata,
        error_handling_type=ERROR_HANDLING_BOTOCORE,
    )


def monkey_patch_botocore_for_xray():
    """Explicit way to monkey-patch botocore to trace AWS API calls."""
    wrapt.wrap_function_wrapper(
        'botocore.client',
        'BaseClient._make_api_call',
        xray_botocore_api_call,
    )


def extract_http_metadata(wrapped, instance, args, kwargs, return_value):
    """Provide HTTP request metadata for improved visualization.

    See documentation for this data structure:
    http://docs.aws.amazon.com/xray/latest/devguide/xray-api-segmentdocuments.html#api-segmentdocuments-http
    """
    response = return_value
    LOGGER.debug(
        'Extracting HTTP metadata',
        args=args,
        kwargs=kwargs,
    )
    if 'request' in kwargs:
        request = kwargs['request']
    else:
        request = args[0]

    metadata = {
        'http': {
            'request': {
                'method': request.method.upper(),
                'url': request.url,
            },
        },
    }
    if response is not None:
        metadata['http']['response'] = {
            'status': response.status_code,
        }

    return metadata


def xray_requests_send(wrapped, instance, args, kwargs):
    """Wrapper around the requests library's low-level send method."""
    return generic_xray_wrapper(
        wrapped, instance, args, kwargs,
        name='requests',
        namespace='remote',
        metadata_extractor=extract_http_metadata,
    )


def monkey_patch_requests_for_xray():
    """Explicit way to monkey-patch requests to trace HTTP requests."""
    wrapt.wrap_function_wrapper(
        'requests.sessions',
        'Session.send',
        xray_requests_send,
    )


def to_safe_annotation_key(key):
    """Xray doesn't like keys that have punctuations
    and likes to silently drop things."""
    safe_key = key.translate(str.maketrans('', '', string.punctuation))
    return safe_key


def to_safe_annotation_value(value):
    """Xray doesn't like values that are not strings."""
    return str(value)



from __future__ import absolute_import

import boto3 as real_boto3
from boto3 import docs, exceptions, logging, resources, session  # noqa: F401
from boto3 import set_stream_logger, setup_default_session  # noqa: F401
from botocore.client import Config

DEFAULT_TIMEOUT = None
DEFAULT_CONNECT_TIMEOUT = None
DEFAULT_READ_TIMEOUT = None


def set_default_timeout(timeout=None, connect_timeout=None, read_timeout=None):
    """
    The purpose of this function is to install default socket timeouts that
    different than those used by boto3 and botocore. Clients obtained from the
    `fleece.boto3.client()` function will have these timeouts set, unless
    explicitly overriden.

    The default timeouts set through this option apply to the entire process.
    For that reason, it is recommended that this function is only called once
    during startup, and from the main thread, before any other threads are
    spawned.

    Socket timeouts are automatically retried by boto, according to its own
    retry policies.

    :param timeout timeout for socket connections and reads in seconds. This is
                   a convenience argument that applies the same default to both
                   connection and read timeouts.
    :param connect_timeout timeout for socket connections in seconds.
    :param read_timeout timeout for socket reads in seconds.
    """
    global DEFAULT_TIMEOUT
    global DEFAULT_CONNECT_TIMEOUT
    global DEFAULT_READ_TIMEOUT
    DEFAULT_TIMEOUT = timeout
    DEFAULT_CONNECT_TIMEOUT = connect_timeout
    DEFAULT_READ_TIMEOUT = read_timeout


def client(*args, **kwargs):
    """
    Create a low-level service client by name using the default session.
    Socket level timeouts are preconfigured according to the defaults set via
    the `fleece.boto3.set_default_timeout()` function, or they can also be set
    explicitly for a client by passing the `timeout`, `connect_timeout` or
    `read_timeout` arguments.
    """
    timeout = kwargs.pop('timeout', DEFAULT_TIMEOUT)
    connect_timeout = kwargs.pop('connect_timeout',
                                 DEFAULT_CONNECT_TIMEOUT or timeout)
    read_timeout = kwargs.pop('read_timeout', DEFAULT_READ_TIMEOUT or timeout)

    config = Config(connect_timeout=connect_timeout,
                    read_timeout=read_timeout)
    return real_boto3.client(*args, config=config, **kwargs)


def resource(*args, **kwargs):
    """
    Create a resource service client by name using the default session.
    Socket level timeouts are preconfigured according to the defaults set via
    the `fleece.boto3.set_default_timeout()` function, or they can also be set
    explicitly for a client by passing the `timeout`, `connect_timeout` or
    `read_timeout` arguments.
    """
    timeout = kwargs.pop('timeout', DEFAULT_TIMEOUT)
    connect_timeout = kwargs.pop('connect_timeout',
                                 DEFAULT_CONNECT_TIMEOUT or timeout)
    read_timeout = kwargs.pop('read_timeout', DEFAULT_READ_TIMEOUT or timeout)

    config = Config(connect_timeout=connect_timeout,
                    read_timeout=read_timeout)
    return real_boto3.resource(*args, config=config, **kwargs)



from cProfile import Profile
try:
    from cStringIO import StringIO
except ImportError:
    # StringIO was moved in Python 3
    from io import StringIO
from functools import wraps
from pstats import Stats
import random
import re

from fleece import log

DEFAULT_LOGGER = log.getLogger('profiler')

RE_SUMMARY_LINE = re.compile(
    r'^\s+(?P<total_calls>\d+) function calls '
    r'\((?P<primitive_calls>\d+) primitive calls\) '
    r'in (?P<total_time>\d+\.\d+) seconds$'
)
RE_PROFILE_LINE = re.compile(
    r'^\s+(?P<ncalls>\d+)\s+'
    r'(?P<tottime>\d+\.\d+)\s+'
    r'(?P<tpercall>\d+\.\d+)\s+'
    r'(?P<cumtime>\d+\.\d+)\s+'
    r'(?P<cpercall>\d+\.\d+)\s+'
    r'(?P<filename>.*):(?P<lineno>\d+)\((?P<function>.*)\)$'
)
# Percentage of calls that should be profiled
PROFILE_SAMPLE = 0.5
# This means that only code that is part of the uploaded Lambda package will
# be present in the profiling report. This will include bundled dependencies,
# but not the standard library, or packages that are provided by Lambda
# (boto*).
DEFAULT_FILTER = ['/var/task/']
DEFAULT_LIMIT = 20


def process_profiling_data(stream, logger, event):
    profiling_data = []
    extra_dict = {}

    raw_string = stream.getvalue()
    lines = raw_string.split('\n')

    # First line contains a summary of the profiled data.
    match_summary = RE_SUMMARY_LINE.match(lines[0])
    if match_summary is not None:
        extra_dict = match_summary.groupdict()
    # In the rest of the data, there might be empty lines, so we have to handle
    # non-matching lines gracefully.
    for line in lines[1:]:
        match = RE_PROFILE_LINE.match(line)
        if match is not None:
            profiling_data.append(match.groupdict())

    logger.info(
        'Profiling completed',
        lambda_event=event,
        profiling_data=profiling_data,
        **extra_dict
    )


def profile_handler(sample=PROFILE_SAMPLE, stats_filter=None,
                    stats_limit=DEFAULT_LIMIT, logger=DEFAULT_LOGGER):

    def decorator(func):

        @wraps(func)
        def wrapper(event, context, *args, **kwargs):
            if random.random() <= PROFILE_SAMPLE:
                print_stats_filter = stats_filter or DEFAULT_FILTER
                print_stats_filter.append(stats_limit)

                profile = Profile()
                profile.enable()
                try:
                    return_value = func(event, context, *args, **kwargs)
                finally:
                    profile.disable()

                    stream = StringIO()
                    stats = Stats(profile, stream=stream)
                    stats.sort_stats('cumulative')
                    stats.print_stats(*print_stats_filter)
                    process_profiling_data(stream, logger, event)
            else:
                logger.info('Skipping profiling')
                return_value = func(event, context, *args, **kwargs)

            return return_value

        return wrapper

    return decorator



import re
import sys


def _fullmatch(pattern, text, *args, **kwargs):
    """re.fullmatch is not available on Python<3.4."""
    match = re.match(pattern, text, *args, **kwargs)
    return match if match.group(0) == text else None


if sys.version_info >= (3, 4):
    # just use the built-in re.fullmatch
    fullmatch = re.fullmatch
else:
    fullmatch = _fullmatch



import sys
import pkg_resources

commands = ['build', 'run', 'config']


def print_help():
    print('Available sub-commands: {}.'.format(', '.join(commands)))
    print('Use "fleece <sub-command> --help" for usage.')


def main():
    if len(sys.argv) == 1:
        print_help()
        sys.exit(0)

    if sys.argv[1] in commands:
        # Check that the CLI dependencies are installed before executing the
        # command.
        deps = pkg_resources.get_distribution('fleece')._dep_map.get(
            'cli', [])
        for dep in deps:
            try:
                # PyYAML really messes this up.
                if dep.project_name == 'PyYAML':
                    __import__('yaml')
                else:
                    __import__(dep.project_name)
            except ImportError:
                print('Dependency "{}" is not installed. Did you run '
                      '"pip install fleece[cli]"?'.format(dep))
                sys.exit(1)

        # execute the command
        module = __import__('fleece.cli.' + sys.argv[1])
        module = getattr(module.cli, sys.argv[1])
        getattr(module, 'main')(sys.argv[2:])
    else:
        if sys.argv[1] not in ['--help', '-h']:
            print('"{}" is not an available fleece sub-command.'.format(
                sys.argv[1]))
        print_help()



import copy
import datetime
import random
import uuid

from fleece.events import format_event


class LambdaContext(object):
    """
    A python class to simulate the context class passed in when a lambda is
    invoked.

    Most functionality can be overridden with a simple sublass.
    Ex:  get_remaining_time_in_millis() simply returns a static value.
     This could be overridden to help test time boxed functions.
    """
    def __init__(self,
                 function_name='test_function',
                 function_version=1,
                 remaining_time_in_milli=9999,
                 memory_limit_in_mb=256,
                 region='us-east-1',
                 account_id='999999999999',
                 stage='test_stage',
                 aws_request_id=None,
                 client_context=None):
        """

        :param function_name: Name of the given function
        :param function_version: version of the function
        :param remaining_time_in_milli: default value tp return for
        get_remaining_time_in_millis
        :param memory_limit_in_mb: Amount of ram the lambda has access to
        :param region: Which region the lambda is executing in
        :param account_id: The account id the lambda is running in
        :param stage: Which "stage" the lambda is running in
        :param aws_request_id: Defaults to a random UUID
        :param client_context: Defaults to None
        """

        # Setup Default values not exposed directly
        self._region = region
        self._account_id = account_id
        self._stage = stage
        self._remaining_time_in_milli = remaining_time_in_milli

        # Set Basic fields
        self.memory_limit_in_mb = memory_limit_in_mb
        self.function_name = function_name
        self.function_version = function_version

        # Use formatters to create appropriate values
        if aws_request_id is None:
            self.aws_request_id = self._generate_aws_request_id()
        else:
            self.aws_request_id = aws_request_id

        if client_context is None:
            self.client_context = self._generate_client_context()

        self.invoked_function_arn = self._generate_function_arn()
        self.log_group_name = self._generate_log_group_name()
        self.log_stream_name = self._generate_log_stream_name()

    def get_remaining_time_in_millis(self):
        return self._remaining_time_in_milli

    def _generate_log_group_name(self):
        log_group_format = '/aws/lambda/{function_name}'
        return log_group_format.format(function_name=self.function_name)

    def _generate_log_stream_name(self):
        stream_format = '{year}/{month}/{day}/[{iterator}]/{random_id}'
        return stream_format.format(
            year=datetime.datetime.year,
            month=datetime.datetime.month,
            day=datetime.datetime.day,
            iterator=str(random.randint(1, 999)),
            random_id=str(uuid.uuid4()).replace('-', '')
        )

    def _generate_aws_request_id(self):
        return str(uuid.uuid4())

    def _generate_client_context(self):
        return {}

    def _generate_function_arn(self):
        arn_format = "arn:aws:lambda:{region}:{account_id}:function:{function_name}:{stage}"  # noqa

        return arn_format.format(
            region=self._region,
            account_id=self._account_id,
            function_name=self.function_name,
            stage=self._stage
        )


class LambdaEvent(object):
    """
    A class the help generating events passed to lambda functions

    After an object has been created use generate() to create events

    To override and of the default values simply pass in a dictionary for the
    appropriate keyword and it will be merged.

    EX:
    event.generate(gateway={"http-method": "post"},
                   header={"x-auth-token": "SOME VALUE"})

    """
    body = {}
    gateway = {
            "http-method": "GET",
            "request-id": str(uuid.uuid4()),
            "resource-path": "/bogus/path/",
            "stage": "test_stage",
            "stage-data": {}
    }
    header = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate",
            "CloudFront-Forwarded-Proto": "https",
            "CloudFront-Is-Desktop-Viewer": "true",
            "CloudFront-Is-Mobile-Viewer": "false",
            "CloudFront-Is-SmartTV-Viewer": "false",
            "CloudFront-Is-Tablet-Viewer": "false",
            "CloudFront-Viewer-Country": "US",
            "Host": "BOGUS.execute-api.us-east-1.amazonaws.com",
            "User-Agent": "python-requests/2.9.1",
            "Via": "1.1 BOGUS.cloudfront.net (CloudFront)",
            "X-Amz-Cf-Id": "FAKE_TID",
            "X-Forwarded-For": "127.0.0.1",
            "X-Forwarded-Port": "443",
            "X-Forwarded-Proto": "https",
            "x-auth-token": "FAKE_TOKEN"
    }
    operation = 'test:operation'
    path = {}
    querystring = {}
    requestor = {
        "account-id": "",
        "api-key": "",
        "caller": "",
        "source-ip": "127.0.0.1",
        "user": "",
        "user-agent": "python-requests/2.9.1",
        "user-arn": ""
    }
    merge_dicts = True

    def __init__(self,
                 body=None,
                 gateway=None,
                 header=None,
                 operation=None,
                 path=None,
                 querystring=None,
                 requestor=None):

        self.body = body or LambdaEvent.body
        self.gateway = gateway or LambdaEvent.gateway
        self.header = header or LambdaEvent.header
        self.operation = operation or LambdaEvent.operation
        self.path = path or LambdaEvent.path
        self.querystring = querystring or LambdaEvent.querystring
        self.requestor = requestor or LambdaEvent.requestor

    def _generate_body(self, merge, kwargs):
        return dict_update(self.body or LambdaEvent.body, merge, kwargs)

    def _generate_gateway(self, merge, kwargs):
        return dict_update(self.gateway or LambdaEvent.gateway, merge, kwargs)

    def _generate_header(self, merge, kwargs):
        return dict_update(self.header or LambdaEvent.header, merge, kwargs)

    def _generate_operation(self, operation=None):
        if operation is not None:
            return operation
        if self.operation is not None:
            return self.operation
        return LambdaEvent.operation

    def _generate_path(self, merge, kwargs):
        return dict_update(self.path or LambdaEvent.path, merge, kwargs)

    def _generate_querystring(self, merge, kwargs):
        return dict_update(self.querystring or LambdaEvent.querystring, merge,
                           kwargs)

    def _generate_requestor(self, merge, kwargs):
        return dict_update(self.querystring or LambdaEvent.requestor, merge,
                           kwargs)

    def generate(self,
                 merge_with_default=True,
                 **kwargs):

        body = self._generate_body(
            merge_with_default,
            kwargs.get('body', {}))
        gateway = self._generate_gateway(
            merge_with_default,
            kwargs.get('gateway', {}))
        header = self._generate_header(
            merge_with_default,
            kwargs.get('header', {}))
        operation = self._generate_operation(kwargs.get('operation', None))
        path = self._generate_path(
            merge_with_default,
            kwargs.get('path', {}))
        querystring = self._generate_querystring(
            merge_with_default,
            kwargs.get('querystring', {}))
        requestor = self._generate_requestor(
            merge_with_default,
            kwargs.get('requestor', {}))

        event = {
            'operation': operation,
            'parameters': {
                'requestor': requestor,
                'request': {
                    'body': body,
                    'path': path,
                    'querystring': querystring,
                    'header': header
                },
                'gateway': gateway
            }
        }

        return copy.deepcopy(event)


class LambdaRequestGenerator(object):
    """
    This class takes a given context and event and simplifies a generating a
    request. This allows for deeper more targeted testing of code without
    having to jump through the full set of routers. Most higher level functions
    should be able to take a request and proceed as expected.

    And kwargs pass to generate_request will be handed off to the event
    generator.

    This allows you to set up sane defaults for your app and replace only the
    bits you want to test on the fly.
    """

    def __init__(self, event=None, context=None):
        self.event = event or LambdaEvent()
        self.context = context or LambdaContext()

    def generate_request(self, **kwargs):
        request = format_event(
            event=self.event.generate(**kwargs),
            context=self.context)
        return request


def dict_update(base, merge, kwargs):
    if not merge:
        return kwargs

    response = base.copy()
    response.update(kwargs)
    return response



#!/usr/bin/env python

import os
import setuptools
import subprocess
import sys


here = os.path.dirname(os.path.realpath(__file__))
about = {}
with open(os.path.join(here, 'fleece', '__about__.py'), 'r') as abt:
    exec(abt.read(), about)


# Add the commit hash to the keywords for sanity.
if any(k in ' '.join(sys.argv).lower() for k in ['upload', 'dist']):
    try:
        current_commit = subprocess.check_output(
            ['git', 'rev-parse', 'HEAD']).decode('utf-8').strip()
    except (OSError, subprocess.CalledProcessError):
        pass
    else:
        if current_commit and len(current_commit) == 40:
            about['__keywords__'].append(current_commit[:8])


# pandoc --from=markdown_github --to=rst README.md --output=README.rst
LONG_DESCRIPTION = ''
readme = os.path.join(here, 'README.rst')
if os.path.isfile(readme):
    with open(os.path.join(here, 'README.rst')) as rdme:
        LONG_DESCRIPTION = rdme.read()


INSTALL_REQUIRES = [
    'structlog>=15.3.0',
    'requests>=2.9.1',
    'boto3>=1.0.0',
    'wrapt>=1.10.10',
]


EXTRAS_REQUIRE = {
    'connexion': [
        'connexion==1.1.9',
        'Flask==0.12.2',
        'Werkzeug==0.15.3',
    ],
    'cli': [
        'docker==3.5.0',
        'PyYAML>=3.12',
        'ruamel.yaml>=0.15.34',
        'six>=1.11.0'
    ],
    'wsgi': [
        'Werkzeug==0.15.3',
    ],
}

ENTRY_POINTS = {
    'console_scripts': [
        'fleece = fleece.cli.main:main',
    ],
}

TESTS_REQUIRE = [
    'coverage>=4.0.3',
    'flake8>=2.5.1',
    'mock>=1.3.0',
    'nose>=1.3.7',
    'pylint>=1.5.4',
]


CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Topic :: Software Development',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
]


package_attributes = {
    'author': about['__author__'],
    'author_email': about['__email__'],
    'classifiers': CLASSIFIERS,
    'description': about['__summary__'],
    'extras_require': EXTRAS_REQUIRE,
    'install_requires': INSTALL_REQUIRES,
    'entry_points': ENTRY_POINTS,
    'keywords': about['__keywords__'],
    'license': about['__license__'],
    'long_description': LONG_DESCRIPTION,
    'name': about['__title__'],
    'packages': setuptools.find_packages(exclude=['tests']),
    'include_package_data':  True,
    'tests_require': TESTS_REQUIRE,
    'test_suite': 'tests',
    'url': about['__url__'],
    'version': about['__version__'],
}

setuptools.setup(**package_attributes)



from __future__ import absolute_import

from requests import *  # noqa
from requests.packages.urllib3.util import Retry
from requests.adapters import HTTPAdapter
from requests import Session as _Session

DEFAULT_CONNECT_TIMEOUT = None
DEFAULT_READ_TIMEOUT = None
DEFAULT_RETRY_ARGS = {}


def set_default_timeout(timeout=None, connect_timeout=None, read_timeout=None):
    """
    The purpose of this function is to install default socket timeouts and
    retry policy for requests calls. Any requests issued through the requests
    wrappers defined in this module will have these automatically set, unless
    explicitly overriden.

    The default timeouts and retries set through this option apply to the
    entire process. For that reason, it is recommended that this function is
    only called once during startup, and from the main thread, before any
    other threads are spawned.

    :param timeout timeout for socket connections and reads in seconds. This is
                   a convenience argument that applies the same default to both
                   connection and read timeouts.
    :param connect_timeout timeout for socket connections in seconds.
    :param read_timeout timeout for socket reads in seconds.
    """
    global DEFAULT_CONNECT_TIMEOUT
    global DEFAULT_READ_TIMEOUT
    DEFAULT_CONNECT_TIMEOUT = connect_timeout if connect_timeout is not None \
        else timeout
    DEFAULT_READ_TIMEOUT = read_timeout if read_timeout is not None \
        else timeout


def set_default_retries(*args, **kwargs):
    """
    This function installs a default retry mechanism to be used in requests
    calls. The arguments are those of urllib3's `Retry` object (see
    http://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry).

    Examples:

        # use 3 retries, for default conditions only (network errors/timeouts)
        set_default_retries(total=3)
        set_default_retries(3)  # total can be given as a positional argument

        # use 5 retries, and add 429 and 503 responses to the list of
        # retryable conditions
        set_default_retries(5, status_forcelist=[429, 503])

        # use 5 retries with an exponential backoff factor of 1
        set_default_retries(5, backoff_factor=1)
    """
    global DEFAULT_RETRY_ARGS
    DEFAULT_RETRY_ARGS = kwargs
    if len(args) > 1:
        ValueError('too many arguments')
    elif len(args) == 1:
        DEFAULT_RETRY_ARGS['total'] = args[0]


class Session(_Session):
    """
    This is a wrapper for requests's `Session` class that adds support for
    injecting a retry mechanism into all outgoing requests.

    :param retries The retry mechanism to use. If `None`, the default retry
                   arguments are used. If an `int`, retry as many times, using
                   the default retry arguments. If a `dict`, use as arguments
                   to a urllib3 `Retry` object. If none of the above types,
                   then it is assumed to be a `urllib3.Retry` instance.
    """
    def __init__(self, timeout=None, retries=None):
        super(Session, self).__init__()
        self.timeout = timeout
        if retries is None:
            retry = Retry(**DEFAULT_RETRY_ARGS)
        elif isinstance(retries, int):
            args = DEFAULT_RETRY_ARGS.copy()
            args.pop('total', None)
            retry = Retry(total=retries, **args)
        elif isinstance(retries, dict):
            retry = Retry(**retries)
        self.mount('http://', HTTPAdapter(max_retries=retry))
        self.mount('https://', HTTPAdapter(max_retries=retry))

    def request(self, method, url, **kwargs):
        """
        Send a request.
        If timeout is not explicitly given, use the default timeouts.
        """
        if 'timeout' not in kwargs:
            if self.timeout is not None:
                kwargs['timeout'] = self.timeout
            else:
                kwargs['timeout'] = (DEFAULT_CONNECT_TIMEOUT,
                                     DEFAULT_READ_TIMEOUT)
        return super(Session, self).request(method=method, url=url, **kwargs)


def request(method, url, **kwargs):
    """
    Wrapper for the `requests.request()` function.
    It accepts the same arguments as the original, plus an optional `retries`
    that overrides the default retry mechanism.
    """
    retries = kwargs.pop('retries', None)
    with Session(retries=retries) as session:
        return session.request(method=method, url=url, **kwargs)


def get(url, params=None, **kwargs):
    """Sends a GET request.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
                   string for the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs.setdefault('allow_redirects', True)
    return request('get', url, params=params, **kwargs)


def options(url, **kwargs):
    """Sends a OPTIONS request.
    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs.setdefault('allow_redirects', True)
    return request('options', url, **kwargs)


def head(url, **kwargs):
    """Sends a HEAD request.
    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    kwargs.setdefault('allow_redirects', False)
    return request('head', url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """Sends a POST request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in
                 the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the
                 :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    return request('post', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    """Sends a PUT request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in
                 the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    return request('put', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    """Sends a PATCH request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in
                 the body of the :class:`Request`.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    return request('patch', url,  data=data, **kwargs)


def delete(url, **kwargs):
    """Sends a DELETE request.
    :param url: URL for the new :class:`Request` object.
    :param \*\*kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """
    return request('delete', url, **kwargs)



#!/usr/bin/env python
import argparse
import os
import subprocess
import sys

import boto3
import requests
import yaml

from fleece import utils


RS_AUTH_ERROR = 'Rackspace authentication failed:\nStatus: {}\nResponse: {}'
ACCT_NOT_FOUND_ERROR = 'No AWS account for `{}` found in config'
NO_USER_OR_APIKEY_ERROR = 'You must provide a Rackspace username and apikey'
NO_STAGE_DATA = 'No stage named "{}" found in config'
NO_ENV_IN_STAGE = 'No default environment defined for stage "{}"'
ENV_AND_ACCT_ERROR = 'Use only ONE of `--environment` or `--account`'
NO_ACCT_OR_ENV_ERROR = 'You must provide either `--environment` or `--account`'
ENV_AND_ROLE_ERROR = ('`--role` cannot be used with `--environment` '
                      '- use `role: <rolename>` in the config file instead')
FAWS_API_URL = ('https://accounts.api.manage.rackspace.com/v0/awsAccounts/{0}'
                '/credentials')
RS_IDENTITY_URL = 'https://identity.api.rackspacecloud.com/v2.0/tokens'
FAWS_API_ERROR = ('Could not fetch AWS Account credentials.\nStatus: {}\n'
                  'Reason: {}')


def parse_args(args):

    for arg in args:
        try:
            cutoff = args.index('--')
            ap_args = args[:cutoff]
            run_args = args[cutoff + 1:]
        except ValueError:
            ap_args = args
            run_args = None

    parser = argparse.ArgumentParser(
        prog='fleece run',
        description=('Run command in environment with AWS credentials from '
                     'Rackspace FAWS API')
    )
    parser.add_argument('--username', '-u', type=str,
                        help=('Rackspace username. Can also be set via '
                              'RS_USERNAME environment variable'))
    parser.add_argument('--apikey', '-k', type=str,
                        help=('Rackspace API key. Can also be set via '
                              'RS_API_KEY environment variable'))
    parser.add_argument('--config', '-c', type=str,
                        default='./environments.yml',
                        help=('Path to YAML config file with defined accounts '
                              'and aliases. Default is ./environments.yml'))
    parser.add_argument('--account', '-a', type=str,
                        help=('AWS account number. Cannot be used with '
                              '`--environment`'))
    parser.add_argument('--environment', '-e', type=str,
                        help=('Environment alias to AWS account defined in '
                              'config file. Cannot be used with `--account`'))
    parser.add_argument('--stage', '-s', type=str,
                        help=('Stage name defined in config file which links '
                              'to an environment. Cannot be used with '
                              '`--account`'))
    parser.add_argument('--role', '-r', type=str,
                        help=('Role name to assume after obtaining credentials'
                              ' from FAWS API'))
    if run_args is None:
        parser.add_argument('command', type=str, help=('Command to execute'))

    args_namespace = parser.parse_args(ap_args)

    if run_args is not None:
        args_namespace.command = run_args

    return args_namespace


def assume_role(credentials, account, role):
    """Use FAWS provided credentials to assume defined role."""
    sts = boto3.client(
        'sts',
        aws_access_key_id=credentials['accessKeyId'],
        aws_secret_access_key=credentials['secretAccessKey'],
        aws_session_token=credentials['sessionToken'],
    )
    resp = sts.assume_role(
        RoleArn='arn:aws:sts::{}:role/{}'.format(account, role),
        RoleSessionName='fleece_assumed_role'
    )
    return {
        'accessKeyId': resp['Credentials']['AccessKeyId'],
        'secretAccessKey': resp['Credentials']['SecretAccessKey'],
        'sessionToken': resp['Credentials']['SessionToken'],
    }


def get_stage_data(stage, data):
    if stage in data:
        return data[stage]
    for s in data:
        if s.startswith('/'):
            if utils.fullmatch(s.split('/')[1], stage):
                return data[s]
    return None


def get_environment(config, stage):
    """Find default environment name in stage."""
    stage_data = get_stage_data(stage, config.get('stages', {}))
    if not stage_data:
        sys.exit(NO_STAGE_DATA.format(stage))
    try:
        return stage_data['environment']
    except KeyError:
        sys.exit(NO_ENV_IN_STAGE.format(stage))


def get_account(config, environment, stage=None):
    """Find environment name in config object and return AWS account."""
    if environment is None and stage:
        environment = get_environment(config, stage)
    account = None
    for env in config.get('environments', []):
        if env.get('name') == environment:
            account = env.get('account')
            role = env.get('role')
            username = os.environ.get(env.get('rs_username_var')) \
                if env.get('rs_username_var') else None
            apikey = os.environ.get(env.get('rs_apikey_var')) \
                if env.get('rs_apikey_var') else None
    if not account:
        sys.exit(ACCT_NOT_FOUND_ERROR.format(environment))
    return account, role, username, apikey


def get_aws_creds(account, tenant, token):
    """Get AWS account credentials to enable access to AWS.

    Returns a time bound set of AWS credentials.
    """
    url = (FAWS_API_URL.format(account))
    headers = {
        'X-Auth-Token': token,
        'X-Tenant-Id': tenant,
    }
    response = requests.post(url, headers=headers,
                             json={'credential': {'duration': '3600'}})

    if not response.ok:
        sys.exit(FAWS_API_ERROR.format(response.status_code, response.text))
    return response.json()['credential']


def get_config(config_file):
    """Get config file and parse YAML into dict."""
    config_path = os.path.abspath(config_file)

    try:
        with open(config_path, 'r') as data:
            config = yaml.safe_load(data)
    except IOError as exc:
        sys.exit(str(exc))

    return config


def get_rackspace_token(username, apikey):
    """Get Rackspace Identity token.

    Login to Rackspace with cloud account and api key from environment vars.
    Returns dict of the token and tenant id.
    """
    auth_params = {
        "auth": {
            "RAX-KSKEY:apiKeyCredentials": {
                "username": username,
                "apiKey": apikey,
            }
        }
    }
    response = requests.post(RS_IDENTITY_URL, json=auth_params)
    if not response.ok:
        sys.exit(RS_AUTH_ERROR.format(response.status_code, response.text))

    identity = response.json()
    return (identity['access']['token']['id'],
            identity['access']['token']['tenant']['id'])


def validate_args(args):
    """Validate command-line arguments."""
    if not any([args.environment, args.stage, args.account]):
        sys.exit(NO_ACCT_OR_ENV_ERROR)
    if args.environment and args.account:
        sys.exit(ENV_AND_ACCT_ERROR)
    if args.environment and args.role:
        sys.exit(ENV_AND_ROLE_ERROR)


def run(args):
    role = args.role

    if args.environment or args.stage:
        config = get_config(args.config)
        account, role, cfg_username, cfg_apikey = get_account(
            config, args.environment, args.stage)
    else:
        cfg_username, cfg_apikey = None, None
        account = args.account

    username = args.username or cfg_username or os.environ.get('RS_USERNAME')
    apikey = args.apikey or cfg_apikey or os.environ.get('RS_API_KEY')
    if not all([username, apikey]):
        sys.exit(NO_USER_OR_APIKEY_ERROR)
    token, tenant = get_rackspace_token(username, apikey)
    faws_credentials = get_aws_creds(account, tenant, token)

    if role:
        aws_credentials = assume_role(faws_credentials, account, role)
    else:
        aws_credentials = faws_credentials

    env = os.environ.copy()
    env['AWS_ACCESS_KEY_ID'] = aws_credentials['accessKeyId']
    env['AWS_SECRET_ACCESS_KEY'] = aws_credentials['secretAccessKey']
    env['AWS_SESSION_TOKEN'] = aws_credentials['sessionToken']

    if isinstance(args.command, list):
        command = ' '.join("'{}'".format(c.replace("'", "\\'"))
                           for c in args.command)
    else:
        command = args.command

    sys.exit(subprocess.call(command, shell=True, env=env))


def main(args):
    parsed_args = parse_args(args)
    validate_args(parsed_args)
    run(parsed_args)



#!/usr/bin/env python
from __future__ import print_function

import argparse
from datetime import datetime
import hashlib
from io import BytesIO
import os
import subprocess
import sys
import tarfile
import tempfile

import docker
from docker import errors

build_dir = os.path.abspath(os.path.dirname(__file__))


def parse_args(args):
    parser = argparse.ArgumentParser(prog='fleece build',
                                     description='Simple Lambda builder.')
    parser.add_argument('--python36', '-3', action='store_true',
                        help='use Python 3.6 (default: Python 2.7)')
    parser.add_argument('--inject-build-info', action='store_true',
                        help='injects config.json into lambda, which contains '
                             'build time and version hash')
    parser.add_argument('--rebuild', action='store_true',
                        help='rebuild Python dependencies')
    parser.add_argument('--requirements', '-r', type=str,
                        default='',
                        help=('requirements.txt file with dependencies '
                              '(default: $service_dir/src/requirements.txt)'))
    parser.add_argument('--pipfile', '-p', type=str,
                        default='',
                        help=('directory containing Pipfile.lock '
                              '(default: $service_dir/Pipfile.lock)'))
    parser.add_argument('--dependencies', '-d', type=str, default='',
                        help='comma separated list of system dependencies')
    parser.add_argument('--target', '-t', type=str,
                        default='',
                        help=('target directory for lambda_function.zip '
                              '(default $service_dir/dist)'))
    parser.add_argument('--source', '-s', type=str,
                        default='',
                        help=('source directory to include in '
                              'lambda_function.zip (default: '
                              '$service_dir/src)'))
    parser.add_argument('--exclude', '-e', type=str, nargs='+',
                        help='glob pattern to exclude')
    parser.add_argument('service_dir', type=str,
                        help=('directory where the service is located '
                              '(default: $pwd)'))
    return parser.parse_args(args)


def get_version_hash():
    sha1 = os.environ.get('CIRCLE_SHA1', None)
    if sha1 is not None:
        return sha1

    try:
        sha1 = subprocess.check_output(
            'git log -1 | head -1 | cut -d" " -f2',
            shell=True,
        ).decode('utf-8').strip()
    except Exception as exc:
        print('Could not determine SHA1: {}'.format(exc))
        sha1 = None

    return sha1


def clean_up_container(container, clean_up_volumes=True):
    try:
        container.remove(v=clean_up_volumes)
    except docker.errors.APIError as e:
        if 'Failed to destroy btrfs snapshot' in str(e):
            # circle-ci builds do not get permissions that allow them to remove
            # containers, see https://circleci.com/docs/1.0/docker-btrfs-error/
            pass
        else:
            raise


def retrieve_archive(container, dist_dir):
    stream, stat = container.get_archive('/dist/lambda_function.zip')
    f = BytesIO()
    for chunk in stream:
        f.write(chunk)
    f.seek(0)
    with tarfile.open(fileobj=f, mode='r') as t:
        t.extractall(path=dist_dir)


def put_files(container, src_dir, path, single_file_name=None):
    stream = BytesIO()

    with tarfile.open(fileobj=stream, mode='w', dereference=True) as tar:
        if single_file_name:
            arcname = single_file_name
        else:
            arcname = os.path.sep
        tar.add(src_dir, arcname=arcname)
    stream.seek(0)
    container.put_archive(data=stream, path=path)


def create_volume(name):
    api = docker.from_env(version='auto')
    api.volumes.create(name)


def destroy_volume(name):
    api = docker.from_env(version='auto')
    try:
        volume = api.volumes.get(name)
    except errors.NotFound:
        return
    try:
        volume.remove()
    except docker.errors.APIError as exc:
        if '409 Client Error' in str(exc):
            print('Unable to remove volume - {}\n{}'.format(name, str(exc)))
        else:
            raise


def create_volume_container(image='alpine:3.4', command='/bin/true', **kwargs):
    api = docker.from_env(version='auto')
    api.images.pull(image)
    container = api.containers.create(
        image,
        command,
        **kwargs
    )
    return container


def build(args):
    python_version = 'python36' if args.python36 else 'python27'
    inject_build_info = args.inject_build_info
    service_dir = os.path.abspath(args.service_dir)
    service_name = os.path.basename(service_dir)

    if args.source:
        src_dir = os.path.abspath(args.source)
    else:
        src_dir = os.path.join(service_dir, 'src')
    if not os.path.exists(src_dir):
        print('Error: src directory not found!')
        sys.exit(1)

    if args.target:
        dist_dir = os.path.abspath(args.target)
    else:
        dist_dir = os.path.join(service_dir, 'dist')
    if not os.path.exists(dist_dir):
        os.makedirs(dist_dir)

    build_cache_dir = os.path.join(service_dir, 'build_cache')
    if not os.path.exists(build_cache_dir):
        os.makedirs(build_cache_dir)

    requirements_path = None
    pipfile = None
    if args.requirements:
        if args.pipfile:
            print('Error: `requirements` and `pipfile` are mutually exclusive')
            sys.exit(1)

        requirements_path = os.path.abspath(args.requirements)
    elif args.pipfile:
        pipfile = args.pipfile
    else:
        potential_req_file = os.path.join(service_dir, 'src/requirements.txt')
        potential_pipfile = os.path.join(service_dir, 'Pipfile.lock')

        if (os.path.exists(potential_pipfile) and
                not os.path.exists(potential_req_file)):
            pipfile = potential_pipfile
        else:
            if os.path.exists(potential_pipfile):
                print('Warning- Pipfile and requirements.txt were found. '
                      'Using requirements.txt. To use the Pipfile, specify '
                      '`--pipfile` or delete the requirements.txt file.')
            # don't worry; if requirements.txt isn't found an error will be
            # raised below
            requirements_path = potential_req_file

    dependencies = args.dependencies.split(',')

    if requirements_path:
        print(requirements_path)
        if not os.path.exists(requirements_path):
            print('Error: requirements file {} not found!'.format(
                requirements_path))
            sys.exit(1)
        _build(service_name=service_name,
               python_version=python_version,
               src_dir=src_dir,
               dependencies=dependencies,
               requirements_path=requirements_path,
               rebuild=args.rebuild,
               exclude=args.exclude,
               dist_dir=dist_dir,
               inject_build_info=inject_build_info)
    else:
        print(pipfile)
        if not os.path.exists(pipfile):
            print('Error: pipfile {} not found!'.format(pipfile))
            sys.exit(1)

        _build_with_pipenv(service_name=service_name,
                           python_version=python_version,
                           src_dir=src_dir,
                           dependencies=dependencies,
                           pipfile=pipfile,
                           rebuild=args.rebuild,
                           exclude=args.exclude,
                           dist_dir=dist_dir,
                           inject_build_info=inject_build_info)

        # If pipfile was specified, we need to write the requirements out
        # to a temporary directory.


def _build_with_pipenv(service_name, python_version, src_dir, pipfile,
                       dependencies, rebuild, exclude, dist_dir,
                       inject_build_info):
    requirements_path = None
    tmpdir = tempfile.mkdtemp()

    # it's too bad Python 2 doesn't have tempfile.TemporaryDirectory :(
    try:
        requirements_path = os.path.join(tmpdir, 'pipfile-requirements.txt')
        print('Creating temporary requirements.txt from Pipenv.lock...')
        requirements_txt_contents = subprocess.check_output(
            'pipenv lock -r',
            shell=True,
            cwd=os.path.dirname(pipfile))

        with open(requirements_path, 'w') as f:
            f.write(requirements_txt_contents)

        _build(service_name=service_name,
               python_version=python_version,
               src_dir=src_dir,
               requirements_path=requirements_path,
               dependencies=dependencies,
               rebuild=rebuild,
               exclude=exclude,
               dist_dir=dist_dir,
               inject_build_info=inject_build_info)
    finally:
        if requirements_path:
            try:
                os.remove(requirements_path)
            except OSError:
                pass
        os.rmdir(tmpdir)


def _build(service_name, python_version, src_dir, requirements_path,
           dependencies, rebuild, exclude, dist_dir, inject_build_info):
    print('Building {} with {}...'.format(service_name, python_version))

    try:
        docker_api = docker.from_env(version='auto')
    except:
        raise RuntimeError("Docker not found.")

    image, _logs = docker_api.images.build(
        path=build_dir, tag=service_name,
        buildargs={'python_version': python_version,
                   'deps': ' '.join(dependencies)})

    with open(requirements_path, 'rb') as fp:
        dependencies_sha1 = hashlib.sha1(fp.read()).hexdigest()

    # Set up volumes
    src_name = '{}-src'.format(service_name)
    req_name = '{}-requirements'.format(service_name)
    dist_name = '{}-dist'.format(service_name)
    create_volume(src_name)
    create_volume(req_name)
    create_volume(dist_name)

    src = create_volume_container(
        volumes=[
            '{}:/src'.format(src_name),
            '{}:/requirements'.format(req_name),
            '{}:/dist'.format(dist_name)]
    )

    # We want our build cache to remain over time if possible.
    build_cache_name = '{}-build_cache'.format(service_name)
    try:
        build_cache = docker_api.containers.get(build_cache_name)
    except errors.NotFound:
        create_volume(build_cache_name)
        build_cache = create_volume_container(
            name=build_cache_name,
            volumes=['{}:/build_cache'.format(build_cache_name)])

    # Inject our source and requirements
    put_files(src, src_dir, '/src')
    put_files(src, requirements_path, '/requirements',
              single_file_name='requirements.txt')

    # Environment variables (including any PIP configuration variables)
    environment = {'DEPENDENCIES_SHA': dependencies_sha1,
                   'VERSION_HASH': get_version_hash(),
                   'BUILD_TIME': datetime.utcnow().isoformat(),
                   'REBUILD_DEPENDENCIES': '1' if rebuild else '0',
                   'EXCLUDE_PATTERNS': ' '.join(
                       ['"{}"'.format(e) for e in exclude or []])}
    for var, value in os.environ.items():
        if var.startswith('PIP_'):
            environment[var] = value

    # Run Builder
    container = docker_api.containers.run(
        command=[
            '/docker_build_lambda.sh',
            'yes' if inject_build_info else ''
        ],
        image=image.tags[0],
        environment=environment,
        volumes_from=[src.id, build_cache.id],
        detach=True)
    for line in container.logs(stream=True, follow=True):
        sys.stdout.write(line.decode('utf-8'))
    status = container.wait()
    exit_code = status.get('StatusCode')
    error_msg = status.get('Error')

    if exit_code or exit_code is None:
        print('Error: build ended with exit code = '
              '{}\nError Message: {}.'.format(exit_code, error_msg))
    else:
        # Pull out our built zip
        retrieve_archive(container, dist_dir)

        # Clean up generated containers
        clean_up_container(container)
        clean_up_container(src)
        destroy_volume(dist_name)
        destroy_volume(req_name)
        destroy_volume(src_name)

        print('Build completed successfully.')
    sys.exit(exit_code)


def main(args):
    build(parse_args(args))
